/*****************************************************************************
 *                                                                           *
 *     Header for memory management                                          *
 *     Version 1.0                        Author      :    Henry Koplien     *
 *                                        Create date :    19.11.98          *
 *                                        Last edit   :    27.11.98          *
 *****************************************************************************/
/**
  *  This class handles the memory management.
  *  The benefit of this class is to keep track of the bytes used. It is
  *  therefore an easy way to find memory leaks. This watchdog is only active
  *  if the switch {\bf MEMCTRL} is set at compile time. C++ gives a handy
  *  solution with {\em operator overloading}. Care should be taken of correct
  *  results. The g++ is known as a buggy compiler. The author posted the
  *  misbehaviour to the GNU-Foundation.
  *
  *  @author  Henry Koplien
  *  @version 1.0 (first release 19. November 1998, last edit 27. November 1998)
  */
class memory {

public:

#ifdef MEMCTRL
  /**
    *  The operator is used to build a new memory management function.
    *  The number of requested bytes is protocoled.
    *
    *  @param  size how many bytes requested to allocate
    *  @return Pointer to the memory where to find the amount of
    *          requested bytes on success, otherwise {\bf NULL}
    */
  void             *operator new         (size_t size);
  /**
    *  The operator is used to build a delete memory management function.
    *  The number of released bytes is protocoled. The defined parameters
    *  are not used, they are for the correct syntax and therefore syntactic 
    *  sugar. What doesn't work is to keep track of used bytes allocated in
    *  case of arrays. There is no mechanism in C++ to handle such things.
    *
    *  @param  pointer_to_bytes  Pointer to allocated memory
    *  @param  size              how many bytes requested at allocate
    */
  void             operator  delete      (void *pointer_to_bytes, size_t size);
#endif
  /**
    *  The function is used to report the final memory releases.
    *  The last five destructor calls are listed including the remaining bytes
    *  which should be released in the near future.
    */
  void             epilog_on             (void);
  /**
    *  Nice function to print out the maximum bytes ever requested.
    *  Till now there are three different functionalities supported.
    *  \begin{itemize}
    *  \item[1] prints out actual number of calls and bytes used, return 0
    *  \item[2] prints out maximum number of bytes ever requested, return 0
    *  \item[3] no readable output but returns the maximum number of bytes ever
    *           requested
    *  \end{itemize}
    *
    *  @param  id defines what to do
    *  @return amount of bytes, kind of bytes depends on {\bf id}
    */
  long             status                (int id);
};
