/*****************************************************************************
 *                                                                           *
 *     Class definition for the serial I/O interface                         *
 *     Version 1.1                        Author      :    Henry Koplien     *
 *                                        Create date :    19.11.98          *
 *                                        Last edit   :    29.11.98          *
 *****************************************************************************/
#include "global.hpp"
#include "mail.hpp"
#include "message.hpp"
#include "chain.hpp"
#include "rs232.hpp"

#define FAIL      -1    // Bad return code
#define IGCHAR    -1    // Ignored char, if not otherwise defined

#define START      0
#define SEND       1
#define RECV       2
#define IDLE       3

#define MAX_RETRY  3    // Number of retries if the communication fails
#define MAX_RING 300    // Seconds till Ring_Alarm become TRUE
#define ALIVE   1800    // Seconds between the alive logs

//
// All the static variables for the signal handler...
//
static int alert_count; // Counter for alarms
static int ring_count;  // Ring counter
static int sigescape;   // Signal flag for escape
static int state;       // State of action
static int old_state;   // Old state of action
static int sigfd;       // File descriptor for signal handler
static int rs232_lines; // Control lines of the RS232
//
// Flags for signal activities
//
static int ioctl_err;   // Flag for ioctl
static int change_sig;  // Changed lines flag
static int rng_on;      // Flag for RNG (Power Failure)
static int rng_off;     // Flag for not RNG
static int state_sta;   // Start state error flag
static int state_rx;    // Receive state error flag
static int state_tx;    // Send state error flag
static int state_idle;  // Idle state flag
static int unknown;     // Flag for unknown signal
static int internal;    // Internal error for unknown state

/***********************************************************************
 *  Name:        Signal-Handler.  (No Method!)                         *
 *  Parameter:   Signal-Number.                                        *
 *  Return:      Nothing.                                              *
 *  Function:    Handles the defined signal.                           *
 ***********************************************************************/
static void signal_handler(int sig) {
int ctrl_lines;

  //
  // Look periodically after the serial line...
  //
  if (ioctl(sigfd, TIOCMGET, &ctrl_lines) == FAIL)
    ioctl_err++;
  else {
    if (rs232_lines != ctrl_lines)
      change_sig = 1;

    //
    // If we have no CTS we stay in (or enter) the START state
    // The only way to escape form the START state is a CTS signal
    //
    if (!(ctrl_lines & TIOCM_CTS))
      state = START;
    else
      if (state == START)
        state = IDLE;
    rs232_lines = ctrl_lines;
  };

  //
  // Look for a "Ring" and log the first one
  // If the Ring disappears, log it too
  //
  if (rs232_lines & TIOCM_RNG) {
    if (ring_count == 0) {
      rng_on++;
      ring_count = 1;
    };
  }
  else
    if (ring_count) {
      rng_off++;
      ring_count = 0;
    };

  //
  // Handle the different signals...
  //
  switch (sig) {
    case SIGALRM: //
                  // Normal time out, first check if we have a ring alarm!
                  //
                  if (rs232_lines & TIOCM_RNG)
                    ring_count++;
                    
                  switch (state) {
                    case START: if (old_state != START)
                                  state_sta = 1;
                                break;

                    case SEND : state_tx = 1;
                                break;

                    case RECV : state_rx = 1;
                                break;

                    case IDLE : if (alert_count > ALIVE) {
                                  state_idle  = 1;
                                  alert_count = 0;
                                };
                                break;

                    default   : internal = 1;
                  };

                  if (internal)
                    alarm(0);
                  else {
                    alert_count++;
                    alarm(1);
                    old_state = state;
                  };
                  break;

    case SIGUSR1: //
                  // Enable the watchdog timer.
                  //
                  alert_count = 0;
                  alarm(1);
                  break;

    case SIGUSR2: //
                  // Disable the watchdog timer.
                  //
                  alarm(0);
                  break;

    case SIGHUP : //
                  // Hang up signal is ignored (for now)
                  //
                  break;

    case SIGINT : //
    case SIGQUIT: // Set sigescape != 0 --> Break Out, but silent please
    case SIGTSTP: //
                  sigescape = 2;
                  break;

    case SIGTERM: //
                  // Terminate signal, set sigescape != 0 (and 1 for a message)
                  //
                  sigescape = 1;
                  break;

    default     : if (sig)
                    unknown = sig;
                  else
                    unknown = 99;
  };
};

/***********************************************************************
 *  Name:        Constructor.                                          *
 *  Parameter:   None.                                                 *
 *  Return:      Nothing.                                              *
 *  Function:    Tries to open the device and prepare it.              *
 ***********************************************************************/
rs232::rs232(void) {
int             ldisc   = NTTYDISC;        // Line Discipline (new vs. old)
struct sgttyb   para    = {  B2400,        // Modes... Receive Rate
                             B2400,        // Transmit Rate
                            IGCHAR,        // Erase-Char (doesn't care)
                            IGCHAR,        // Kill-Char (doesnt't care)
                               RAW };      // Raw mode (8 bit)
int             flush   = 0;               // Flush In and Out
struct tchars   special = { IGCHAR,        // Interrupt char
                            IGCHAR,        // Quit char
                            IGCHAR,        // Start char
                            IGCHAR,        // Stop char
                            IGCHAR,        // End of file
                            IGCHAR };      // Input delimiter
int             local   = LPASS8 | LNOFLSH;// Local mode
struct ltchars  locspec = { IGCHAR,        // Stop process signal char
                            IGCHAR,        // Delayed stop process signal char
                            IGCHAR,        // Reprint line char
                            IGCHAR,        // Flush output char
                            IGCHAR,        // Word erase char
                            IGCHAR };      // Literal next char
char   device[20], apid[20];
char   lock_file[100], fake_file[100];
char   *fptr;
int    lockfd, pid, length;
int    ctrl_lines;
string text;

  sprintf(device, "/dev/%s", SERIAL_DEVICE);
  sprintf(lock_file, "%s%s", LOCK_PREFIX, SERIAL_DEVICE);
  sprintf(fake_file, "%sUPS.XXXXXX", LOCK_PREFIX);

  //
  // First open a temp file and ignore the umask
  //
  locked = FALSE;
  if ((lockfd = creat(fptr = mktemp(fake_file), 0644)) == FAIL)
    error(LOCK_FILE, "RS232-Init ( Can't open temp file )");
  (void) chmod(fptr, 0644);
  
  //
  // Write the binary form, we know pid_t is int...
  //
  pid = getpid();
  if (write(lockfd, &pid, sizeof(pid)) != sizeof(pid)) {
    (void) close(lockfd);
    (void) unlink(fptr);
    error(LOCK_FILE, "RS232-Init ( Can't write pid to (temp) lock )");
  };
  (void) close(lockfd);

  //
  // Our file exists, now we try to link it the right file name and drop
  // the temp file
  //
  while (::link(fptr, lock_file) == FAIL)
    if (errno == EEXIST)
      if ((lockfd = open(lock_file, O_RDONLY)) == FAIL) {
        if (errno != ENOENT) {
          (void) unlink(fptr);
          error(LOCK_FILE, "RS232-Init ( Can't open existing lock file )");
        };
      }
      else {
        if ((length = ::read(lockfd, apid, sizeof(apid)-1)) == FAIL) {
          if (errno != ENOENT) {
            (void) close(lockfd);
            (void) unlink(fptr);
            error(LOCK_FILE, "RS232-Init ( Can't read lock file )");
          };
          (void) close(lockfd);
        }
        else {
          (void) close(lockfd);

          if (length != sizeof(pid)) {
            (void) unlink(fptr);
            error(LOCK_FILE, "RS232-Init ( Lock file has wrong size )");
          };

          pid = *( (int *)apid );

          if (pid == getpid()) {
            (void) unlink(fptr);
            error(LOCK_FILE, "RS232-Init ( *WE* own the lock file? )");
          }
          else {
            if ((kill(pid, 0) == FAIL) && errno == ESRCH) {
              if (unlink(lock_file) == FAIL && errno != ENOENT) {
                (void) unlink(fptr);
                error(LOCK_FILE, "RS232-Init ( Can't remove staled lock file (permissions?) )");
              };
              warning("Removed staled lock file...");
            }
            else {
              (void) unlink(fptr);
              error(LOCK_FILE, "RS232-Init ( Device is locked by another process )");
            };
          }; 
        };
      }
    else {
      (void) unlink(fptr);
      error(LOCK_FILE, "RS232-Init ( Can't link to lock file )");
    };
  (void) unlink(fptr);
  locked = TRUE;

  if ((fd = open(device, O_RDWR)) < 0) {
    remove_lock();
    switch (errno) {
      case EACCES      : error(OPEN_SIO, "RS232-Init ( No permissions to access the device )");
      case EFAULT      : error(OPEN_SIO, "RS232-Init ( The path argument points out of process memory space )");
      case EINTR       : error(OPEN_SIO, "RS232-Init ( Got interrupt signal )");
      case EIO         : error(OPEN_SIO, "RS232-Init ( I/O error )");
      case EISDIR      : error(OPEN_SIO, "RS232-Init ( Requested device is a directory )");
      case ELOOP       : error(OPEN_SIO, "RS232-Init ( Too many symbolic links )");
      case EMFILE      : error(OPEN_SIO, "RS232-Init ( Ran out of system file descriptors )");
      case ENAMETOOLONG: error(OPEN_SIO, "RS232-Init ( Device name too long )");
      case ENFILE      : error(OPEN_SIO, "RS232-Init ( No more system device entries possible )");
      case ENOENT      : error(OPEN_SIO, "RS232-Init ( Non existent device )");
      case ENOTDIR     : error(OPEN_SIO, "RS232-Init ( Missing directory in device name )");
      case ENXIO       : error(OPEN_SIO, "RS232-Init ( Special file format not supported )");
      case EOPNOTSUPP  : error(OPEN_SIO, "RS232-Init ( Operation on sockets not allowed )");
      case EROFS       : error(OPEN_SIO, "RS232-Init ( File system is read only (i.e. the hard disk) )");
      case ETXTBSY     : error(OPEN_SIO, "RS232-Init ( The device is a shared text file and still open )");
      default          : error(INTERNAL, "RS232-Init ( Got an unsupported open ioctl return )");
    };
  };

  if (ioctl(fd, TIOCSETD,  &ldisc) == FAIL)
    error(OPEN_SIO, "RS232-Init ( Writing the line discipline )");
  if (ioctl(fd, TIOCSETP,  &para) == FAIL)
    error(OPEN_SIO, "RS232-Init ( Writing the parameters )");
  if (ioctl(fd, TIOCFLUSH, &flush) == FAIL)
    error(OPEN_SIO, "RS232-Init ( Flushing the line )");
  if (ioctl(fd, TIOCSETC,  &special) == FAIL)
    error(OPEN_SIO, "RS232-Init ( Writing the special characters )");
  if (ioctl(fd, TIOCLSET,  &local) == FAIL)
    error(OPEN_SIO, "RS232-Init ( Writing the local mode )");
  if (ioctl(fd, TIOCSLTC,  &locspec) == FAIL)
    error(OPEN_SIO, "RS232-Init ( Writing the local special characters )");
  if (ioctl(fd, TIOCMGET, &ctrl_lines) == FAIL)
    error(OPEN_SIO, "RS232-Init ( Reading the control lines )");
  ctrl_lines &= ~TIOCM_RTS;
  if (ioctl(fd, TIOCMSET, &ctrl_lines) == FAIL)
    error(OPEN_SIO, "RS232-Init ( Writing the control lines )");

  sprintf(text, "RS232 successfully initialized, control lines are %s%s%s%s%s%s%s",
                (ctrl_lines & TIOCM_LE)?"LE ":"",
                (ctrl_lines & TIOCM_DTR)?"DTR ":"",
                (ctrl_lines & TIOCM_RTS)?"RTS ":"",
                (ctrl_lines & TIOCM_CTS)?"CTS ":"",
                (ctrl_lines & TIOCM_CAR)?"DCD ":"",
                (ctrl_lines & TIOCM_RNG)?"RNG ":"",
                (ctrl_lines & TIOCM_DSR)?"DSR":"");
  message(text);

  (void) signal(SIGHUP,  signal_handler);
  (void) signal(SIGINT,  signal_handler);
  (void) signal(SIGQUIT, signal_handler);
  (void) signal(SIGTSTP, signal_handler);
  (void) signal(SIGTERM, signal_handler);
  (void) signal(SIGALRM, signal_handler);
  (void) signal(SIGUSR1, signal_handler);
  (void) signal(SIGUSR2, signal_handler);

  sigfd       = fd;
  rs232_lines = ctrl_lines;
  ring_count  = 0;
  sigescape   = 0;
  ioctl_err   = 0;
  change_sig  = 0;
  rng_on      = 0;
  rng_off     = 0;
  state_sta   = 0;
  state_rx    = 0;
  state_tx    = 0;
  state_idle  = 0;
  unknown     = 0;
  internal    = 0;
  state       = START;
  old_state   = IDLE;
  enable_watchdog();
};

/***********************************************************************
 *  Name:        Destructor.                                           *
 *  Parameter:   None.                                                 *
 *  Return:      Nothing.                                              *
 *  Function:    Close the device and release the lock.                *
 ***********************************************************************/
rs232::~rs232(void) {
  disable_watchdog();
  (void) close(fd);
  remove_lock();
  (void) signal(SIGHUP,  SIG_IGNORE);
  (void) signal(SIGINT,  SIG_IGNORE);
  (void) signal(SIGQUIT, SIG_IGNORE);
  (void) signal(SIGTSTP, SIG_IGNORE);
  (void) signal(SIGTERM, SIG_IGNORE);
  (void) signal(SIGALRM, SIG_IGNORE);
  (void) signal(SIGUSR1, SIG_IGNORE);
  (void) signal(SIGUSR2, SIG_IGNORE);
  message("RS232 successfully brought down");
};

/***********************************************************************
 *  Name:        remove_lock.                                          *
 *  Parameter:   None.                                                 *
 *  Return:      Nothing.                                              *
 *  Function:    Remove the lock file, if any.                         *
 ***********************************************************************/
void rs232::remove_lock(void) {
char lock_file[100];

  if (locked) {
    sprintf(lock_file, "%s%s", LOCK_PREFIX, SERIAL_DEVICE);
    if (unlink(lock_file) == FAIL)
      error(LOCK_FILE, "RS232-Remove ( Can't remove lock file )");
  };
  locked = FALSE;
};

/***********************************************************************
 *  Name:        RTS_toggle.                                           *
 *  Parameter:   None.                                                 *
 *  Return:      Nothing.                                              *
 *  Function:    Enable RTS for 5 seconds.                             *
 ***********************************************************************/
void rs232::RTS_toggle(void) {
int ctrl_lines;

  if (ioctl(fd, TIOCMGET, &ctrl_lines) == FAIL)
    error(OPEN_SIO, "RS232-Toggle ( Reading the control lines )");
  if (!(ctrl_lines & TIOCM_RTS)) {
    ctrl_lines |= TIOCM_RTS;
    if (ioctl(fd, TIOCMSET, &ctrl_lines) == FAIL)
      error(OPEN_SIO, "RS232-Toggle ( Writing the control lines )");
    message("RTS-Toggle -> RTS High");
    sleep(5);
    ctrl_lines &= ~TIOCM_RTS;
    if (ioctl(fd, TIOCMSET, &ctrl_lines) == FAIL)
      error(OPEN_SIO, "RS232-Toggle ( Writing the control lines )");
    message("RTS-Toggle -> RTS Low again");
  };
};

/***********************************************************************
 *  Name:        Ring_Alarm.                                           *
 *  Parameter:   None.                                                 *
 *  Return:      TRUE or FALSE.                                        *
 *  Function:    Returns TRUE if the ring counter exceeds a limit or   *
 *               the battery is low.                                   *
 ***********************************************************************/
bool rs232::Ring_Alarm(void) {
int ctrl_lines;

  if (ioctl(fd, TIOCMGET, &ctrl_lines) == FAIL)
    error(OPEN_SIO, "RS232-Ring ( Reading the control lines )");
  if ((ring_count >= MAX_RING || (ctrl_lines & TIOCM_CAR)) && (ctrl_lines & TIOCM_RNG))
    return TRUE;
  return FALSE;
};

/***********************************************************************
 *  Name:        Signal_Escape.                                        *
 *  Parameter:   None.                                                 *
 *  Return:      TRUE or FALSE.                                        *
 *  Function:    Returns TRUE if an escape signal was recognized.      *
 ***********************************************************************/
bool rs232::Signal_Escape(void) {
  if (sigescape) {
    if (sigescape == 1) { 
      message("Going down on signal 15");
      sigescape = 2;
    };
    return TRUE;
  };
  return FALSE;
};

/***********************************************************************
 *  Name:        Activities.                                           *
 *  Parameter:   None.                                                 *
 *  Return:      Nothing.                                              *
 *  Function:    Record the activities since last call.                *
 ***********************************************************************/
void rs232::Activities(void) {
string text;

  if (ioctl_err) {
    sprintf(text, "( %d time(s) since last call )", ioctl_err);
    error(IOCTL_ERR, text);
    ioctl_err = 0;
  };

  if (change_sig) {
    sprintf(text, "RS232 changed the control lines to %s%s%s%s%s%s%s",
                   (rs232_lines & TIOCM_LE)?"LE ":"",
                   (rs232_lines & TIOCM_DTR)?"DTR ":"",
                   (rs232_lines & TIOCM_RTS)?"RTS ":"",
                   (rs232_lines & TIOCM_CTS)?"CTS ":"",
                   (rs232_lines & TIOCM_CAR)?"DCD ":"",
                   (rs232_lines & TIOCM_RNG)?"RNG ":"",
                   (rs232_lines & TIOCM_DSR)?"DSR":"");
    message(text);
    change_sig = 0;
  };

  if (rng_on) {
    sprintf(text, "( %d time(s) since last call )", rng_on);
    error(RING_ON, text);
    rng_on = 0;
  };

  if (rng_off) {
    sprintf(text, "( %d time(s) since last call )", rng_off);
    error(RING_OFF, text);
    rng_off = 0;
  };

  if (internal)
    error(INTERNAL, "Signal_Handler");

  if (state_sta) {
    warning("Data connection timed out, no Clear to Send (CTS)!");
    state_sta = 0;
  };

  if (unknown) {
    sprintf(text, "Signal handler received a suspect signal %d (ignored)", unknown);
    warning(text);
    unknown = 0;
  };

  if (state_rx) {
    message("Time out signal while receiving");
    state_rx = 0;
  };

  if (state_tx) {
    message("Time out signal while sending");
    state_tx = 0;
  };

  if (state_idle) {
    message("Daemon alive...");
    state_idle = 0;
  };
};

/***********************************************************************
 *  Name:        enable_watchdog.                                      *
 *  Parameter:   None.                                                 *
 *  Return:      Nothing.                                              *
 *  Function:    Enables sniffing the interface.                       *
 ***********************************************************************/
void rs232::enable_watchdog(void) {
  (void) kill(getpid(), SIGUSR1);
};

/***********************************************************************
 *  Name:        disable_watchdog.                                     *
 *  Parameter:   None.                                                 *
 *  Return:      Nothing.                                              *
 *  Function:    Disables sniffing the interface.                      *
 ***********************************************************************/
void rs232::disable_watchdog(void) {
  (void) kill(getpid(), SIGUSR2);
};

/***********************************************************************
 *  Name:        crc.                                                  *
 *  Parameter:   1. Crc value.                                         *
 *               2. Data.                                              *
 *  Return:      New crc value.                                        *
 *  Function:    Use data to build the new crc checksum.               *
 ***********************************************************************/
unsigned short int rs232::crc(unsigned short int crc_value, unsigned char data) {
unsigned short int crc_temp;

  crc_value ^= (data << 8) & 0xff00;
  crc_temp   = (((crc_value >> 4) ^ crc_value) >> 4) & 0x0ff0;
  crc_value ^= crc_temp;
  crc_value ^= (crc_temp >> 7) | (crc_temp << 9);
  return (crc_value >> 8) | (crc_value << 8);
};

/***********************************************************************
 *  Name:        tx_crc_init.                                          *
 *  Parameter:   None.                                                 *
 *  Return:      Nothing.                                              *
 *  Function:    Prepares crc checksum for sending.                    *
 ***********************************************************************/
void rs232::tx_crc_init(void) {
  tx_crc = TX_CRC_INIT;
};

/***********************************************************************
 *  Name:        rx_crc_init.                                          *
 *  Parameter:   None.                                                 *
 *  Return:      Nothing.                                              *
 *  Function:    Prepares crc checksum for receiving.                  *
 ***********************************************************************/
void rs232::rx_crc_init(void) {
  rx_crc = RX_CRC_INIT;
};

/***********************************************************************
 *  Name:        rx_ok.                                                *
 *  Parameter:   None.                                                 *
 *  Return:      CRC status TRUE or FALSE.                             *
 *  Function:    Uses crc checksum as an error indicator.              *
 ***********************************************************************/
bool rs232::rx_ok(void) {
  return !rx_crc;
};

/***********************************************************************
 *  Name:        send.                                                 *
 *  Parameter:   Data.                                                 *
 *  Return:      TRUE or FALSE.                                        *
 *  Function:    Sends one byte. TRUE on success, otherwise FALSE.     *
 ***********************************************************************/
bool rs232::send(const unsigned char data) {
unsigned char buf[1];

  enable_watchdog();
  while (state != IDLE)
    sleep(5);
  state = SEND;

  buf[0] = data;
  tx_crc = crc(tx_crc, data);
  if (write(fd, buf, 1) < 1) {
    remove_lock();
    switch (errno) {
      case EAGAIN     : error(WRITE_SIO, "RS232-Send (  Non blocking device would be delayed on write )");
      case EBADF      : error(WRITE_SIO, "RS232-Send (  File descriptor is not a valid descriptor open for writing )");
      case EPIPE      : error(WRITE_SIO, "RS232-Send ( An attempt is made to write to a pipe that is not open for reading )");
      case EFBIG      : error(WRITE_SIO, "RS232-Send ( An attempt was made to write a file that exceeds the process's file size )");
      case EFAULT     : error(WRITE_SIO, "RS232-Send ( File descriptor points out of memory space )");
      case EINVAL     : error(WRITE_SIO, "RS232-Send ( Negative file descriptor )");
      case ENOSPC     : error(WRITE_SIO, "RS232-Send ( There is no free space remaining on the file system )");
      case EINTR      : error(WRITE_SIO, "RS232-Send ( Got interrupt signal )");
      case EDQUOT     : error(WRITE_SIO, "RS232-Send ( The user's quota of disk blocks has been exhausted )");
      case EIO        : error(WRITE_SIO, "RS232-Send ( I/O error )");
      case EWOULDBLOCK: error(WRITE_SIO, "RS232-Send ( Non blocking write but no data writable )");
      default         : error(INTERNAL,  "RS232-Send ( Got an unsupported write ioctl return )");
    };
  };
  state = IDLE;
  return TRUE;
};

/***********************************************************************
 *  Name:        send_crc.                                             *
 *  Parameter:   None.                                                 *
 *  Return:      TRUE or FALSE.                                        *
 *  Function:    Sends the crc. TRUE on success, otherwise FALSE.      *
 ***********************************************************************/
bool rs232::send_crc(void) {
unsigned char buf[2];

  enable_watchdog();
  while (state != IDLE)
    sleep(5);
  state = SEND;

  buf[0] = tx_crc >> 8;
  buf[1] = tx_crc & 0xff;
  if (write(fd, buf, 2) < 2) {
    remove_lock();
    switch (errno) {
      case EAGAIN     : error(WRITE_SIO, "RS232-CRC (  Non blocking device would be delayed on write )");
      case EBADF      : error(WRITE_SIO, "RS232-CRC (  File descriptor is not a valid descriptor open for writing )");
      case EPIPE      : error(WRITE_SIO, "RS232-CRC ( An attempt is made to write to a pipe that is not open for reading )");
      case EFBIG      : error(WRITE_SIO, "RS232-CRC ( An attempt was made to write a file that exceeds the process's file size )");
      case EFAULT     : error(WRITE_SIO, "RS232-CRC ( File descriptor points out of memory space )");
      case EINVAL     : error(WRITE_SIO, "RS232-CRC ( Negative file descriptor )");
      case ENOSPC     : error(WRITE_SIO, "RS232-CRC ( There is no free space remaining on the file system )");
      case EINTR      : error(WRITE_SIO, "RS232-CRC ( Got interrupt signal )");
      case EDQUOT     : error(WRITE_SIO, "RS232-CRC ( The user's quota of disk blocks has been exhausted )");
      case EIO        : error(WRITE_SIO, "RS232-CRC ( I/O error )");
      case EWOULDBLOCK: error(WRITE_SIO, "RS232-CRC ( Non blocking write but no data writable )");
      default         : error(INTERNAL,  "RS232-CRC ( Got an unsupported write ioctl return )");
    };
  };
  state = IDLE;
  return TRUE;
};

/***********************************************************************
 *  Name:        receive.                                              *
 *  Parameter:   1. Pointer to char buffer.                            *
 *               2. Number of requested bytes.                         *
 *  Return:      TRUE or FALSE.                                        *
 *  Function:    Reads n bytes, waits till completion or error.        *
 ***********************************************************************/
bool rs232::receive(unsigned char *cptr, int n) {
unsigned char rb[100], uc;
bool          missing;
int           count, i;

  enable_watchdog();
  while (state != IDLE)
    sleep(5);
  state = RECV;

  count = 0;
  if (n > 0) {
    missing = TRUE;
    while (missing) {
      if (available() >= n)
        missing = FALSE;
      else {
        count = ::read(fd, rb, 100);
        if (count != FAIL) {
          if (count)
            if (!add(rb, count)) {
              remove_lock();
              error(CANT_ADD, "RS232-Receive");
            };
        }
        else {
          if (errno != EINTR)
            remove_lock();
          switch (errno) {
            case EBADF      : error(READ_SIO, "RS232-Receive ( Operation on sockets not allowed )");
            case EFAULT     : error(READ_SIO, "RS232-Receive ( File descriptor points out of memory space )");
            case EIO        : error(READ_SIO, "RS232-Receive ( I/O error )");
            case EINTR      : if (alert_count > MAX_RETRY) {
                                state = IDLE;
                                return FALSE;
                              }
                              else
                                if (available() != lseek(fd, available(), SEEK_SET)) {
                                  remove_lock();
                                  error(READ_SIO, "RS232-Receive ( Got interrupt signal )");
                                };
                              break;
            case EINVAL     : error(READ_SIO, "RS232-Receive ( Negative file descriptor or tried to read more bytes than available )");
            case EWOULDBLOCK: error(READ_SIO, "RS232-Receive ( Non blocking read but no data )");
            default         : error(INTERNAL, "RS232-Receive ( Got an unsupported read ioctl return )");
          };
        };
      };
    };

    for(i = 0; i < n; i++) {
      if (!read(&uc)) {
        remove_lock();
        error(INTERNAL, "RS232-Receive ( Missing existing data )");
      };
      rx_crc = crc(rx_crc, uc);
      *cptr++ = uc;
    };

    state = IDLE;
    return TRUE;
  };

  state = IDLE;
  return FALSE;
};
