/*****************************************************************************
 *                                                                           *
 *     Unix interface to the goodies...                                      *
 *     Version 1.1                        Author      :    Henry Koplien     *
 *                                        Create date :    19.11.98          *
 *                                        Last edit   :    29.11.98          *
 *****************************************************************************/
#include "main.hpp"

/**
  *  @name            The UPS Daemon for APC's Back-PRO    
  *
  *  \begin{center}
  *  {\Large A daemon to work with the Uninterruptible Power Supply\\
  *                     from APC (e.g. Back-PRO)\\
  *  }
  *  \end{center}
  *
  *  This package is build to support the uninterruptible power supply from
  *  APC, specifically the UPS Back-PRO. The daemon, once started, disappears
  *  in the background and look for power failures. If a power fail is
  *  detected for a longer period of time, approx. 5 min, the daemon shuts
  *  down the system. If the whole system is on battery, the time may be
  *  shortened if the battery is low. All these activities are logged in the
  *  system log file.\\
  *
  *  \begin{center}
  *  {\Large Have secure times!}
  *  \end{center}
  *
  *  @version       1.1 (last release 29. November 1998)
  *  @memo     A daemon for NeXT serving the Back-PRO from APC
  *  @author               Henry Koplien
  */
//@{

/**
  *  @name                        UPS Back-PRO with NeXT
  *
  *  It is planned to implement the UPS-Link-Language in the future. It is then
  *  possible to read out much more information than power failure or battery
  *  low. It is also possible to set the timings of the shutdown and much more.
  *  It is also possible to track the voltage, over and under voltage, etc.
  *  Additional things that may come are more user friendly start ups. For
  *  instance switches for the daemon for the log interval time, the maximal
  *  run time on battery before a shutdown is invoked, and other parameters.
  *
  *  @author  Henry Koplien
  *  @version 1.1
  *  @memo    A description of the daemon
  */
//@{

//@Include: global.dxx

/**
  *  @name                  main
  *
  *  Interface to the goodies.
  *  At the moment the given parameters are disregarded. This may change in
  *  the future. The functionality of this routine is to fork the started
  *  process and continue in the background. This is a little bit tricky
  *  because we wrote the stuff in C++, which makes a lot of stuff when a
  *  variable is instantiated. We have also problems after the instantiation
  *  with lock files and pending signals. To avoid all these problems we must
  *  dynamically instantiate the objects when running in the back. Otherwise
  *  we have different pid's in the system log and all the bad things mentioned
  *  above. After the set-up we only wait for kill signals or a power failure.
  *  On a power failure the system will shutdown. To accomplish this task the
  *  package have to be started as root. This is normally done with an entry
  *  in the rc.local file. Known problems are an incorrect creation of the lock
  *  file if started too early. A work-around is to wait some seconds before 
  *  starting the daemon.
  * 
  *  @param  arg  number of arguments
  *  @param  argc pointer to the parameters (if any)
  *  @return exit code depends on daemon bring down
  *  @memo   The entry for the daemon
  */
int main(int arg, char **argc) {
/******************************/
static rs232 *PowerConnect;
int          temp_err;

  //
  // Ok, we are going in the background.
  // We have to be aware that the pid changes when we declare
  // an object as a variable, because the started program instantiate
  // this object (e.g. PowerConnect) at once. This means that
  // the constructors are called and log some things in the
  // syslog before the program runs as a daemon in the back with a
  // different child pid. This problem must be avoided because pending
  // signals are canceled and lock files are ignored when forking a
  // process. So the right solution is instantiating the PowerConnect
  // object in the child code...
  //
  switch (fork()) {
    case 0 : //
             // Everything is fine, we are the child...
             //
             sleep(30);
             if ((PowerConnect = new rs232) == NULL)
               exit(-1);
             PowerConnect->message("Daemon is running in the back...");
             break;

    case -1: //
             // So sorry, and save the error number...
             // (The PowerConnect constructor does a lot of things...)
             // The code after the switch is never reached, normally
             //
             temp_err = errno;
             if ((PowerConnect = new rs232) == NULL)
               exit(-1);
             switch (temp_err) {
               case EAGAIN: PowerConnect->error(FORK, "MAIN ( Not enough processes )");
               case ENOMEM: PowerConnect->error(FORK, "MAIN ( Insufficient swap space )");
               default    : PowerConnect->error(INTERNAL, "MAIN ( Got an unsupported fork return )");
             };
             delete PowerConnect;
             return -1;

    default: //
             // We are the parent and we can return if the object pointer is static,
             // otherwise me must exit to keep the local variables.
             //
             return 0;
  };

  //
  // So, stuff is set up, let's wait for a power failure
  //
  while (!PowerConnect->Ring_Alarm() && !PowerConnect->Signal_Escape()) {

    //
    // Do some things (e.g. Read out data of the UPS, if we ever
    // will get the UPS Link-Language)...
    //
    PowerConnect->Activities();

    sleep(1);
  };

  //
  // Look for a daemon escape signal
  //
  if (PowerConnect->Signal_Escape()) {
    delete PowerConnect;
    return 0;
  };

  //
  // OK, we are dead...
  //
  PowerConnect->disable_watchdog();
  PowerConnect->RTS_toggle();

  //
  // Don't get stopped for some signals and reboot as root...
  // (UPS_REBOOT is the only error call coming back!)
  //
  if (getuid() == 0) {
    (void) signal(SIGHUP,  SIG_IGNORE);
    (void) signal(SIGINT,  SIG_IGNORE);
    (void) signal(SIGQUIT, SIG_IGNORE);
    (void) signal(SIGTSTP, SIG_IGNORE);
    (void) signal(SIGTERM, SIG_IGNORE);
    PowerConnect->error(UPS_REBOOT, "");
    delete PowerConnect;
    reboot(RB_HALT);
  }
  else
    delete PowerConnect;

  return -1;
};

/**
  *  @name                   Low-Level Services
  *
  *  In this section you will find a description of the services which are
  *  needed for proper functioning like the standarized I/O for tty's,
  *  handling of errors, using mail and handling direct hardware I/O or
  *  serial interfaces.
  *
  *  @memo Low level functioning for the upsd
  */
//@{
//@Include: memory.dxx
//@Include: mail.dxx
//@Include: message.dxx
//@Include: chain.dxx
//@Include: rs232.dxx
//@} 

//@}
